import com.cyberbotics.webots.Controller;
import javax.swing.*;
import java.awt.*;
import java.awt.image.*;
import java.lang.Math;
import java.util.*;

public abstract class MyBot extends Controller 
{
  static int leftSensor,rightSensor,camera; // DeviceTags - initialize in reset()
  static MemoryImageSource m_imgSource;
  
  static int m_speed=0, m_rotSpeed=0;   // current traslational and rotational speed
  static int leftSpeed=0, rightSpeed=0; // current left and right speed
  static int old_left=0, old_right=0;   // last encoder value
  static double m_ds, m_da;             // distance traveled & orientation change in a single step
  static double m_ids=0, m_ida=0;       // incremental change
  static double m_x=-445.5, m_y=441.8, m_a=0; // initial position
  // relative to webots coordinates the x axis is flipped and orientation also

  static int m_camWidth, m_camHeight;   // camera width and height, set in enableSensors()
  static void update() {}
  static Map m_map;
  
  public static void reset() // {{{
  { 
    leftSensor=robot_get_device("ds1");
    rightSensor=robot_get_device("ds4");
    camera=robot_get_device("camera");
    System.out.println("MyBot: reset");
  }  // }}}
  public static void enableSensors() // {{{
  {
    robot_battery_sensor_enable(64);
    differential_wheels_enable_encoders(64);
    distance_sensor_enable(leftSensor,64);
    distance_sensor_enable(rightSensor,64);
    camera_enable(camera,64);
    m_camWidth = camera_get_width(camera);
    m_camHeight = camera_get_height(camera);
  } // }}}
  static void updateSpeeds() // {{{
  {
    // differential_wheels_set_speed() uses 0.1 rad/s
    // wheel radius is 8mm, axle length is 53mm
    // translational speed step is 0.8 mm/s, rotational ss is 0.038 rad/s
    leftSpeed = m_speed + m_rotSpeed ;
    rightSpeed = m_speed - m_rotSpeed ;
    differential_wheels_set_speed(leftSpeed,rightSpeed);
    System.out.println("MyBot: ls: "+leftSpeed+" rs: "+rightSpeed);
    if (leftSpeed > 20) System.out.println("MyBot: left speed out of range!!!");
    if (rightSpeed > 20) System.out.println("MyBot: right speed out of range!!!");
  }
  public static void setSpeed(int in_speed) { m_speed = in_speed; updateSpeeds(); }
  static void setRotSpeed(int in_speed) { m_rotSpeed = in_speed; updateSpeeds(); } // }}}
  static void updateRelativePosition() // {{{
  {
    // differential_wheels_get_left_encoder() step is 0.1 deg
    // wheel radius is 8mm
    // one rotation of the wheel coresponds to 2*PI*r distance and 3600 encoder steps
    // one encoder step is then (2*PI*r) / 3600
    // total traveled distance is then (num_steps*PI) / (3600/2/8)
    double left = ((differential_wheels_get_left_encoder() - old_left)*Math.PI)/225;
    double right = ((differential_wheels_get_right_encoder() - old_right)*Math.PI)/225;
    // traveled distance
    m_ds = (left + right)/2;
    // change in orientation
    m_da = Math.asin((left - right)/2/53)*2;
    m_ids += Math.abs(m_ds);
    m_ida += Math.abs(m_da);
    old_left = differential_wheels_get_left_encoder();
    old_right = differential_wheels_get_right_encoder();

    m_x += m_ds*Math.cos(m_a);
    m_y += m_ds*Math.sin(m_a);
    m_a += m_da;
    if (m_a > Math.PI) m_a -= 2*Math.PI;         // to wrap around the m_a
    else if (m_a < -Math.PI) m_a += 2*Math.PI;
  } // }}}
  public static void main() // {{{
  {
    int[] image;
    enableSensors();

    m_imgSource = 
    new MemoryImageSource(
          camera_get_width(camera),
				  camera_get_height(camera), 
          camera_get_image(camera),0, 
          camera_get_width(camera));
    m_imgSource.setAnimated(true);
    m_map = new Map();
	  
    robot_step(64);
    Gui gui = new Gui();
    for(;;)
    {
      image = camera_get_image(camera);
      m_imgSource.newPixels(image, ColorModel.getRGBdefault(), 0, m_camWidth);
      updateRelativePosition();
      gui.update();
      robot_step(64);
    }
  } // }}}
}

class GreenFilter extends RGBImageFilter // {{{
{
  public GreenFilter() { canFilterIndexColorModel = true; }
  public int filterRGB(int x, int y, int rgb) 
  {
    int red, green, blue;
    red   = (rgb & 0x00ff0000)>>16;
    green = (rgb & 0x0000ff00)>>8;
    blue  = (rgb & 0x000000ff);
    if (green>2*red && green>2*blue && green>110) 
      return rgb & 0xff00ff00;
    else 
      return 0xff000000; // black
  }
} // }}}
class GeneralFilter extends RGBImageFilter // {{{
{
  static int s_blue = Color.blue.getRGB(), // colors subtituted into the image 
    s_green = Color.green.getRGB(),        // during filtering
    s_orange = Color.orange.getRGB(),
    s_black = Color.black.getRGB();
  public GeneralFilter() { canFilterIndexColorModel = true; }
  public int filterRGB(int x, int y, int in_rgb) 
  {
    int red, green, blue;
    red   = (in_rgb & 0x00ff0000)>>16;
    green = (in_rgb & 0x0000ff00)>>8;
    blue  = (in_rgb & 0x000000ff);
    if (blue > 80 && blue+20 > 2*red && blue+20 > 2*green) // Wall
      return s_blue;
    else if (green > 210 && red < 60 && blue < 60) // Charged feeder
      return s_green;
    else if (red > 125 && green > 80 && blue == 0) // yellowish
      return s_orange;
    else 
      return s_black;
  }
} // }}}

